/**
  Lighting demo application.
  
  Orion Sky Lawlor, olawlor@acm.org, 2008-04-08 (Public Domain)
*/
#include <GL/glew.h> /*<- for gl...ARB extentions.  Must call glewInit after glutCreateWindow! */
#include <iostream>
#include <GL/glut.h>
#include <math.h>
#include "osl/vec4.h" /*<- handy little vector class */
#include "osl/mat4.h"
using osl::mat4;
/* Convert a struct with .x, .y, .z fields into three arguments */
#define VEC3_TO_XYZ(v) (v).x,(v).y,(v).z


#include "ogl/image.h" /* (originally from) Nigel Stewart's GLT image library */
#include "ogl/image.cpp"

#include "ogl/glsl.h" /* Orion's tiny GLSL library */
#include "ogl/glsl.cpp"

#include "GL/glui.h" /* OpenGL User Interface library (for making a GUI) */

void oglExit(const char *what,const char *why) {
	printf(what,why);
	exit(1);
}
int gi_res=16; /* pixels on a side for little global illumination views */
int gi_self=1; /* show object while computing its own GI views */
float walking_speed=3.0, gi_fov=120.0;
float obj_reflectance=0.8;
float obj_scale=1.0;
int move_camera=1, show_gi=0;
mat4 potato_gui=mat4(1.0);
mat4 light_direction=mat4(1.0);

/******************* Object, read from .obj file ***************/
class obj_mesh {
public:
	obj_mesh(const char *fName);
	
	void draw(void);
	
	void gi(const osl::mat4 &to_World);
private:
	struct myVertex {
		vec3 xyz; /* location */
		vec3 color; /* GI-derived color */
		myVertex(const vec3 &v) :xyz(v), color(0,1,0) {}
	};
	std::vector<myVertex> vtx; /* 3D location of vertices (0-based) */
	GLuint vb, eb; /* GPU vertex buffer object, element buffer object */
	void vb_upload(void); /* copy vtx array data into vb */
	
	std::vector<vec3> vtx_norm; /* GI vertex normals */
	
	/* By normal index */
	std::vector<vec3> norm; /* surface normal (unit length) */
	
	struct myFace {
		int v[3]; /* 1-based vertex index */
		int n[3]; /* 1-based normal vector index */
	};
	std::vector<myFace> tri;
};

obj_mesh::obj_mesh(const char *fName)
{
	FILE *f=fopen(fName,"r");
	char line[1000], stuff[100];
	while (NULL!=fgets(line,999,f)) {
		switch (line[0]) {
		case 'v': { /* Vertex line, like "v 0.1 0.7 0.9" */
			double x,y,z;
			if (4!=sscanf(line,"%99s %lf%lf%lf",stuff,&x,&y,&z)) 
				oglExit("Unparsable v-line '%s'\n",line);
			if (stuff==std::string("v"))
				vtx.push_back(vec3(x,y,z));
			else if (stuff==std::string("vn"))
				norm.push_back(normalize(vec3(x,y,z)));
			else
				oglExit("Unknown v-line '%s'\n",line);
			break;
		}
		case 'f': { /* Face line, like "f 1 2 3 4" or "f 1//1 2//2 3//3 4//4". */
			myFace q;
			for (int i=0;i<3;i++) q.n[i]=0;
			/* FIXME: this is hideous coding, right here... */
			/* Tri face, with normals */
			if (7==sscanf(line,"%99s %d//%d  %d//%d  %d//%d ",stuff,
				&q.v[0],&q.n[0], 
				&q.v[1],&q.n[1], 
				&q.v[2],&q.n[2])
			)
			{
				vtx_norm.resize(vtx.size());
				for (int i=0;i<3;i++)
					vtx_norm[q.v[i]-1]=norm[q.n[i]-1];
				tri.push_back(q);
			}
			/* Tri face, without normals */
			else if (4==sscanf(line,"%99s %d  %d  %d ",stuff,
					&q.v[0], 
					&q.v[1], 
					&q.v[2])
			)
				tri.push_back(q);
			else /* all parse attempts failed! */
				oglExit("Unknown f-line '%s'\n",line);
			break;
		}
		
		default: break; /* ignore unrecognized lines */
		};
	}
	fclose(f);
	
	glGenBuffersARB(1,&vb); 
	vb_upload();
	
	std::vector<unsigned int> tri_idx;
	for (unsigned int t=0;t<tri.size();t++) {
		for (int i=0;i<3;i++) 
			tri_idx.push_back(tri[t].v[i]-1);
	}
	glGenBuffersARB(1,&eb); 
	glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB,eb);
	glBufferDataARB(GL_ELEMENT_ARRAY_BUFFER_ARB,sizeof(tri_idx[0])*tri_idx.size(),
		&tri_idx[0],GL_STATIC_DRAW_ARB);
	glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB,0);
}
void obj_mesh::vb_upload(void) 
{ /* copy vtx array data into vb */
	glBindBufferARB(GL_ARRAY_BUFFER_ARB,vb);
	glBufferDataARB(GL_ARRAY_BUFFER_ARB,sizeof(vtx[0])*vtx.size(),
		&vtx[0],GL_STREAM_DRAW_ARB);
	glVertexPointer(3,GL_FLOAT,sizeof(vtx[0]), (void *)0); /* myVertex.xyz is first thing in struct */
	glColorPointer (3,GL_FLOAT,sizeof(vtx[0]), (void *)12); /* myVertex.color starts 12 bytes after struct start */
	glBindBufferARB(GL_ARRAY_BUFFER_ARB,0);
}
void obj_mesh::draw(void)
{
	/* Turn everything on */
	glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB,eb);
	glBindBufferARB(GL_ARRAY_BUFFER_ARB,vb);
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_COLOR_ARRAY);

	/* Draw our (GPU) triangles with (GPU) vertices */
	glDrawElements(GL_TRIANGLES,tri.size()*3,GL_UNSIGNED_INT,0);

	/* Turn everything off afterwards */
	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_COLOR_ARRAY);
	glBindBufferARB(GL_ARRAY_BUFFER_ARB,0);
	glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB,0);
}
/*********************** App Code **********************/

/* Orthonormal coordinate frame */
class ortho_frame {
public:
	/* Unit vectors pointing along axes of our frame.
		X cross Y is Z. (right-handed coordinate system)
	*/
	vec3 x,y,z;
	ortho_frame() :x(1,0,0), y(0,1,0), z(0,0,1) {}
	
	/* Reorient this coordinate frame by this far 
	   along in the X and Y axes.  "dx" is the distance
	   along the z axis to push the x axis; "dy" 
	   the distance along the z axis to push the y axis.
	*/
	void nudge(double dx,double dy) {
		x+=dx*z;
		y+=dy*z;
		orthonormalize();
	}
	
	/* Reconstruct orthonormal frame from X and Y axes.
	  Y is primary, X is secondary, Z is tertiary.
	*/
	void orthonormalize(void) {
		y=normalize(y);
		z=normalize(cross(x,y));
		x=normalize(cross(y,z));
	}
};

/* Orientation of our camera. Camera looks down +Z axis; Y points down. */
ortho_frame camera_frame;

/* Location of our camera */
vec3 camera_location=vec3(0,-3,1.7);

/* Corresponding key is down */
bool key_is_down[256]={false};

/* Time it took to draw the last frame (for updating physics) */
float seconds_per_frame=0.01;

/* Nice handy timer: returns seconds */
double nice_timer(void) {return glutGet(GLUT_ELAPSED_TIME)*1.0e-3;}

/* Return a random float on [-1,+1] */
float rand_float(void) {
	return (rand()&0xffff)*(1.0/0xffff)*2.0-1.0;
}

/* Return a random two-vector on [-1,+1] */
vec3 rand_vec2(void) {
	return vec3(rand_float(),rand_float(),0.0);
}
/* Return the fractional part of this vector */
vec3 fract(const vec3 &v) {
	return v-vec3(floor(v.x),floor(v.y),floor(v.z));
}

/**
 Stores an entire OpenGL scene.
*/
class the_world {
public:
	the_world();
	
	/* Precompute values needed for a given frame.  May trash matrices! */
	void setup(void);
	
	/* Render world to the current modelview and projection matrices */
	void render(bool for_camera);
private:
	GLuint monkeyTex; /* used by world plane */
	int lit_sz; /* resolution of plane-lighting */
	GLuint planeLit; /* lighting on the ground plane */
	vec3 planeLitScale, planeLitOrigin; /* ground plane lighting positions */
	
	obj_mesh potato; /* our main model */
	mat4 potatoM; /* potato-to-world coordinate shift */
};


/// Make a new empty OpenGL texture with this size and internal format.
GLuint make_tex(int w,int h,GLenum texFormat) {
	GLuint tex;
	glGenTextures(1, &tex);
	glBindTexture(GL_TEXTURE_2D, tex);
	/* turn off mipmapping */
	GLenum filt=GL_LINEAR;
	if (texFormat==GL_RGBA_FLOAT16_ATI) filt=GL_NEAREST; /* or else nVidia 6200 fails below */
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, filt);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, filt);
	glTexImage2D(GL_TEXTURE_2D, 0, texFormat, w, h, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
	return tex;
}


the_world::the_world()
	:lit_sz(8), planeLit(make_tex(lit_sz,lit_sz,GL_RGBA8)),
	 potato("potato.obj")
{
	planeLitScale=vec3(6,6,0);
	planeLitOrigin=-planeLitScale*0.5;

	glGenTextures(1,&monkeyTex); /* make texture handle */

	glBindTexture(GL_TEXTURE_2D,monkeyTex);
	readTextureFromFile("monkey.bmp");
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);

	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAX_ANISOTROPY_EXT,16);

	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,
		GL_LINEAR_MIPMAP_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);

	glTexParameteri(GL_TEXTURE_2D,GL_GENERATE_MIPMAP_SGIS, GL_TRUE);

	glActiveTexture(GL_TEXTURE3); /* plug monkey into 3rd texture unit */
	glBindTexture(GL_TEXTURE_2D,monkeyTex);
	
	glActiveTexture(GL_TEXTURE0); /* Everything else into 0th texture unit */
}

/* Render the world.  Assumes matrices are already set up. */
void the_world::render(bool from_camera) {
/* Clear the sky */
	glClear(GL_COLOR_BUFFER_BIT + GL_DEPTH_BUFFER_BIT);

	static GLhandleARB prog=makeProgramObject(
	"// GLSL Vertex shader\n"
	"varying vec4 light;\n"
	"varying vec3 worldCoords;\n"
	"\n"
	"void main(void) {\n"
	"	light = gl_Color;\n"
	"	worldCoords = vec3( gl_ModelViewMatrix * gl_Vertex );\n"
	"	gl_Position = gl_ProjectionMatrix * gl_ModelViewMatrix * gl_Vertex; \n"
	"}\n"
	,
	"// GLSL Fragment shader\n"
	"varying vec4 light;\n"
	"varying vec3 worldCoords;\n"
	"uniform sampler2D myMonkey;\n"
	"uniform sampler2D planeLit;\n"
	"uniform vec3 planeLitOrigin,planeLitInv;\n"
	"\n"
	"void main(void) {\n"
	"	vec2 litCoords=vec2((worldCoords-planeLitOrigin)*planeLitInv);\n"
	"	vec4 light=texture2D(planeLit,litCoords);\n"
	"	gl_FragColor = 0.4*texture2D(myMonkey,vec2(worldCoords))*light;\n"
	"}\n"
	);
	glUseProgramObjectARB(prog);
	static int set_the_monkey=0;
	if (!set_the_monkey) { set_the_monkey=1;
		glUniform1iARB(glGetUniformLocationARB(prog,"myMonkey"),3);
		glUniform1iARB(glGetUniformLocationARB(prog,"planeLit"),0);
		glUniform3fvARB(glGetUniformLocationARB(prog,"planeLitOrigin"),1,planeLitOrigin);
		glUniform3fvARB(glGetUniformLocationARB(prog,"planeLitInv"),1,
			vec3(1.0/planeLitScale.x,1.0/planeLitScale.y,0.0));
	}
	
	/* Draw the giant ground plane */
	glBegin (GL_TRIANGLE_FAN);
	float z=0.0;
        glVertex3d(0-0.5e3,0-0.5e3,z); 
        glVertex3d(1e3-0.5e3,0-0.5e3,z); 
        glVertex3d(1e3-0.5e3,1e3-0.5e3,z); 
        glVertex3d(0-0.5e3,1e3-0.5e3,z); 
        glEnd();
	
	/* Back to fixed-function OpenGL */
	glUseProgramObjectARB(0);
	
	if (!from_camera && !gi_self) return; /*<- don't draw object in its own images */
	
	/* Draw the potato, with GI vertex colors */
	glLoadMatrixf(&potatoM[0][0]);
	potato.draw();
	glLoadIdentity();
}

the_world *world=0; /* world model (can't initialize it until OpenGL is set up!) */

/* Onscreen location of little global illumination textures */
static int render_x=0,render_y=0;
	
/* Return the amount of light arriving at a point facing this direction.

   WARNING: Trashes all the matrices, and the framebuffer!
*/
vec3 light_calc(vec3 p,vec3 n)
{
/* Set up camera pointing out from this point */
	vec3 y=normalize(cross(vec3(0,1,0),n));

	glMatrixMode(GL_PROJECTION); /* set up projection matrix */
	glLoadIdentity(); /* reset matrix at start of frame */
	gluPerspective(gi_fov,1.0,0.01,1000.0);
	gluLookAt(
		VEC3_TO_XYZ(p), /* camera */
		VEC3_TO_XYZ(p+n), /* looking at this point */
		VEC3_TO_XYZ(y)
	);
	glMatrixMode(GL_MODELVIEW); /* set up model and view matrix */
	glLoadIdentity(); /* reset matrix at start of frame */

/* Set up framebuffer to render a small region */
	int sz=gi_res;
	render_x+=sz;
	if (render_x>=512) {
		render_x=0; 
		render_y+=sz;
		if (render_y>=512) render_y=0;
	}
	
	glViewport(render_x,render_y,sz,sz);
	glScissor(render_x,render_y,sz,sz);
	glEnable(GL_SCISSOR_TEST);
	
/* Render everything */
	world->render(false);

/* Rendered the world, now read back average color.
	FIXME: do averaging on GPU, and read back colors in big batches.
*/
	unsigned char *pix=new unsigned char[sz*sz*3];
	glPixelStorei(GL_PACK_ALIGNMENT,1); /* allow byte alignment */
	glReadPixels(render_x,render_y,sz,sz,GL_RGB,GL_UNSIGNED_BYTE,pix);
	vec3 sum=vec3(0.0);
	for (int y=0;y<sz;y++) 
	for (int x=0;x<sz;x++) {
		sum[0]+=pix[3*(y*sz+x)+0];
		sum[1]+=pix[3*(y*sz+x)+1];
		sum[2]+=pix[3*(y*sz+x)+2];
	}
	delete[] pix;
	return sum*(1.0/(sz*sz))*(1.0/255);
}

/* Do global illumination calculation for this mesh, 
  oriented according to this matrix. */
void obj_mesh::gi(const mat4 &toWorld)
{
	/* Figure out how much light is coming in at each mesh vertex */
	for (unsigned int v=0;v<vtx.size();v++) {
		vec3 p=vec3(toWorld*vec4(vtx[v].xyz,1.0));
		vec3 n=vec3(toWorld*vec4(vtx_norm[v],0.0));
		vec3 reflectance(obj_reflectance);/* this vertex's reflectance */
		vtx[v].color=light_calc(p,n)*reflectance; 
	}
	
	/* Now that we know mesh colors, update the GPU's copy of the mesh */
	vb_upload();
}

/* Precompute values needed for a given frame */
void the_world::setup(void)
{
	/* Capture the potato matrix */
	glMatrixMode(GL_MODELVIEW); /* set up model and view matrix */
	glLoadIdentity(); /* reset matrix at start of frame */
	glTranslatef(0,0.0,1.4); /* float potato in front of user */
	glScalef(obj_scale,obj_scale,obj_scale);
	if (potato_gui==mat4(1.0)) 
	{ /* User hasn't touched controls--rotate slowly */
		glRotatef(nice_timer()*1.3,1.0,0.0,0); /* adjust potato axis */
		glRotatef(nice_timer()*10,0.0,0.0,1); /* slowly rotate potato */
	}
	glRotatef(90.0,1.0,0,0); /* rotate potato to face user */
	glMultMatrixf(&potato_gui[0][0]);
	glGetFloatv(GL_MODELVIEW_MATRIX, &potatoM[0][0]);
	
	glutPostRedisplay(); //<- for potato animation
	
	/* Do potato global illumination */
	potato.gi(potatoM);
	
	/* Do ground-plane global illumination, and bind to texture */
	vec3 *lit=new vec3[lit_sz*lit_sz];
	double s=1.0/lit_sz;
	for (int y=0;y<lit_sz;y++)
	for (int x=0;x<lit_sz;x++) {
		vec3 p=vec3((x+0.5)*s,(y+0.5)*s,0);
		lit[y*lit_sz+x]=light_calc(planeLitScale*p+planeLitOrigin,
			vec3(0,0,1)); /* plane normal and origin */
	}
	glBindTexture(GL_TEXTURE_2D,planeLit);
	glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA8,lit_sz,lit_sz, 0,
		GL_RGB,GL_FLOAT,&lit[0]);
	delete[] lit;
}

void my_display(void) {
	glFinish();
	int last_msec=glutGet(GLUT_ELAPSED_TIME);
	
	glClearColor(0.3,0.5,0.7,0); /* set background color to blue */
	if (!world) world=new the_world();
	
/* Do global illumination stuff */
	render_x=render_y=0;

	world->setup();
	
	/* Show the GI stuff */
	if (show_gi) { 
		glutSwapBuffers();
		return;
	}
	
/* Reset matrices after GI stuff */
	int w=glutGet(GLUT_WINDOW_WIDTH), h=glutGet(GLUT_WINDOW_HEIGHT);
	glViewport(0,0,w,h);
	glScissor(0,0,w,h);
	
	glMatrixMode(GL_MODELVIEW); /* set up model and view matrix */
	glLoadIdentity(); /* reset matrix at start of frame */
		
	glMatrixMode(GL_PROJECTION); /* set up projection matrix */
	glLoadIdentity(); /* reset matrix at start of frame */
	
	
/* Move the camera */
	float dt=seconds_per_frame;/* time between physics updates, seconds */
	
	float vel=walking_speed; /* velocity, in m/s */
	float forward_dir=0.0, strafe_dir=0.0;
	if (key_is_down[(int)'w']) {forward_dir+=1.0;}
	if (key_is_down[(int)'s']) {forward_dir-=1.0;}
	if (key_is_down[(int)'a']) {strafe_dir-=1.0;}
	if (key_is_down[(int)'d']) {strafe_dir+=1.0;}
	
	vec3 camera_vel = vel*dt*(
		camera_frame.z*forward_dir+
		camera_frame.x*strafe_dir
	);
	if (move_camera) 
		camera_location += camera_vel;

	glutPostRedisplay();
	
	/* You are walking: */
	camera_location.z=1.7; /* you can't fly */
	
	camera_frame.x.z=0.0; /* you can't tilt your head */
	camera_frame.y=cross(camera_frame.z,camera_frame.x);
	
	if (false && camera_location.z<1.0) 
	{ /* flight mode: bounce off terrain*/
		camera_location.z=1.0;
		camera_frame.nudge(0,-0.5); /* pull back hard */
	}
	
	camera_frame.orthonormalize();
	
/* Set up camera */
	gluPerspective(90.0, /* <- Y camera field-of-view, in degrees */
		glutGet(GLUT_WINDOW_WIDTH)/(float)glutGet(GLUT_WINDOW_HEIGHT), /* aspect ratio: width over height */
		0.01, /* Near clipping plane depth */
		1000.0 /* Far clipping plane depth */
	);
	gluLookAt(VEC3_TO_XYZ(camera_location),
		VEC3_TO_XYZ(camera_location + camera_frame.z), /* look down +Z */
		VEC3_TO_XYZ(-camera_frame.y) /* OpenGL +Y points up, our Y points down */
	);
	
	glMatrixMode(GL_MODELVIEW); /* set up model and view matrix */
	glLoadIdentity(); /* reset matrix at start of frame */
	
	glClear(GL_COLOR_BUFFER_BIT + GL_DEPTH_BUFFER_BIT);
	glEnable(GL_DEPTH_TEST);
	

/* Draw the world */
	world->render(true);
	
	glutSwapBuffers(); /* copy back buffer to front */
	
/* Statistics */
	glFinish();
	int cur_msec=glutGet(GLUT_ELAPSED_TIME); /* milliseconds since glutInit */
	int msec_elapsed=cur_msec-last_msec;
	printf("%d msec/frame; %.2f fps\n",
		msec_elapsed,1.0e3/msec_elapsed);
	last_msec=cur_msec;
	seconds_per_frame=msec_elapsed*1.0e-3;
}

int mouse_x, mouse_y;
void my_mouse(int state,int buttons,int x,int y) {
	mouse_x=x; mouse_y=y;  /* just stash motion */
}
void my_motion(int x,int y) { 
	float dx=x-mouse_x, dy=y-mouse_y; /* relative motion */
	
	float speed=1.0/100; /* radians rotation per pixel */
	camera_frame.nudge(speed*dx,speed*dy);
	
	mouse_x=x; mouse_y=y; 
	glutPostRedisplay();
}

void my_keyboard(unsigned char key,int mouse_x,int mouse_y) 
{
	//printf("Key '%c' going down\n",key);
	key_is_down[key]=true;
	glutPostRedisplay();
}
void my_keyboardup(unsigned char key,int mouse_x,int mouse_y) 
{
	//printf("Key '%c' going back up\n",key);
	key_is_down[key]=false;
}
int main(int argc,char *argv[]) {
	/* Start camera in reasonable orientation */
	camera_frame.y=vec3(0,0,-1); camera_frame.z=vec3(0,1,0);
	camera_frame.orthonormalize();

	glutInitWindowSize(800,600);
	glutInit(&argc,argv);
	int mode=GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH;
        glutInitDisplayMode (mode);
	int win=glutCreateWindow("Interactive Global Illumination Demo (try the WASD keys)");
	glewInit();
	
	GLUI *glui = GLUI_Master.create_glui_subwindow(win, GLUI_SUBWINDOW_TOP);
	glui->set_main_gfx_window( win );
	
	new GLUI_Checkbox(glui, "Show GI Array?", &show_gi);
	new GLUI_Spinner(glui, "GI FOV: ", &gi_fov);
	new GLUI_Spinner(glui, "GI Res: ", &gi_res);
	
	new GLUI_Column(glui); /* for layout */
	new GLUI_Checkbox(glui, "Object Self-lit", &gi_self);
	new GLUI_Spinner(glui, "Reflectance: ", &obj_reflectance);
	new GLUI_Spinner(glui, "Size: ", &obj_scale);
	//new GLUI_Spinner(glui, "Walking Speed: ", &walking_speed);
	//new GLUI_Spinner(glui, "Potato Count: ", &nPotato);
	//new GLUI_Checkbox(glui, "Move camera?", &move_camera);
	
	new GLUI_Column(glui); /* for layout */
	new GLUI_Rotation(glui,"Object",&potato_gui[0][0]);
	
	glutDisplayFunc(my_display);
	glutMotionFunc(my_motion);
	glutMouseFunc(my_mouse);
	glutKeyboardFunc(my_keyboard);
	glutKeyboardUpFunc(my_keyboardup);
	
	glutMainLoop(); /* processes events */
	return 0;
}
